package org.nissley.bluetooth_tester;
/**
 * Created by John Nissley on 12/25/2016.
 */

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.Context;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.os.ParcelUuid;
import android.util.Log;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

/**
 *  This class does all the work for setting up and managing Bluetooth
 *  outbound connections.
 */
public class BluetoothService {
    //Debugging
    private static final String TAG = "BluetoothTester";

    // Member fields
    private final BluetoothAdapter btAdapter;
    private final Handler btHandler;
    private ConnectThread btConnectThread;
    private ConnectedThread btConnectedThread;
    private int btState;

    // Constants that indicate the current connection state
    public static final int STATE_NONE = 10;       // we're doing nothing
    public static final int STATE_CONNECTING = 11; // now initiating an outgoing connection
    public static final int STATE_CONNECTED = 12;  // now connected to a remote device
    public static final int STATE_LOST_CONNECTION = 13; // lost connection to remote device
    public static final int STATE_CONNECTION_FAILED = 14; // lost connection to remote device

    /**
     * Constructor. Prepares a new BluetoothChat session.
     *
     * @param context The UI Activity Context
     * @param handler A Handler to send messages back to the UI Activity
     */
    public BluetoothService(Context context, Handler handler) {
        btAdapter = BluetoothAdapter.getDefaultAdapter();
        btState = STATE_NONE;
        btHandler = handler;
    }

    /**
     * Set the current state of the chat connection
     *
     * @param state An integer defining the current connection state
     */
    private synchronized void setState(int state) {
        Log.d(TAG, "setState() " + btState + " -> " + state);
        btState = state;

        // Give the new state to the Handler so the UI Activity can update
        btHandler.obtainMessage(Constants.MESSAGE_STATE_CHANGE, state, -1).sendToTarget();
    }

    /**
     * Return the current connection state.
     */
    public synchronized int getState() {
        return btState;
    }

    /**
     * Start the ConnectThread to initiate a connection to a remote device.
     *
     * @param device The BluetoothDevice to connect
     */
    public synchronized void connect(BluetoothDevice device) {
        Log.d(TAG, "connect to: " + device);

        // Cancel any thread attempting to make a connection
        if (btState == STATE_CONNECTING) {
            if (btConnectThread != null) {
                btConnectThread.cancel();
                btConnectThread = null;
            }
        }

        // Cancel any thread currently running a connection
        if (btConnectedThread != null) {
            btConnectedThread.cancel();
            btConnectedThread = null;
        }

        // Start the thread to connect with the given device
        btConnectThread = new ConnectThread(device);
        btConnectThread.start();
        setState(STATE_CONNECTING);
    }

    /**
     * Start the ConnectedThread to begin managing a Bluetooth connection
     *
     * @param socket The BluetoothSocket on which the connection was made
     * @param device The BluetoothDevice that has been connected
     */
    public synchronized void connected(BluetoothSocket socket, BluetoothDevice
            device) {
        Log.d(TAG, "Bluetooth socket connected.");

        // Cancel the thread that completed the connection
        if (btConnectThread != null) {
            btConnectThread.cancel();
            btConnectThread = null;
        }

        // Cancel any thread currently running a connection
        if (btConnectedThread != null) {
            btConnectedThread.cancel();
            btConnectedThread = null;
        }

        // Start the thread to manage the connection and perform transmissions
        btConnectedThread = new ConnectedThread(socket);
        setState(STATE_CONNECTED);
        btConnectedThread.start();

        // Send the name of the connected device back to the UI Activity
        Message msg = btHandler.obtainMessage(Constants.MESSAGE_DEVICE_NAME);
        Bundle bundle = new Bundle();
        bundle.putString(Constants.DEVICE_NAME, device.getName());
        msg.setData(bundle);
        btHandler.sendMessage(msg);
    }

    /**
     * Stop all threads
     */
    public synchronized void stop() {
        Log.d(TAG, "stop");
        setState(STATE_NONE);
        if (btConnectThread != null) {
            btConnectThread.cancel();
            btConnectThread = null;
        }

        if (btConnectedThread != null) {
            btConnectedThread.cancel();
            btConnectedThread = null;
        }
    }

    /**
     * Write to the ConnectedThread in an unsynchronized manner
     *
     * @param out The bytes to write
     * @see ConnectedThread#write(byte[])
     */
    public void write(byte[] out) {
        // Create temporary object
        ConnectedThread r;
        // Synchronize a copy of the ConnectedThread
        synchronized (this) {
            if (btState != STATE_CONNECTED) return;
            r = btConnectedThread;
        }
        // Perform the write unsynchronized
        r.write(out);
    }

    /**
     * Indicate that the connection attempt failed and notify the UI Activity.
     */
    private void connectionFailed() {
        // Send a failure message back to the Activity
        setState(STATE_CONNECTION_FAILED);
        //Message msg = btHandler.obtainMessage(Constants.MESSAGE_TOAST);
        //Bundle bundle = new Bundle();
        //bundle.putString(Constants.TOAST, "Unable to connect device");
        //msg.setData(bundle);
        //btHandler.sendMessage(msg);
    }

    /**
     * Indicate that the connection was lost and notify the UI Activity.
     */
    private void connectionLost() {
        // Send a failure message back to the Activity
        setState(STATE_LOST_CONNECTION);
        //Message msg = btHandler.obtainMessage(Constants.MESSAGE_TOAST);
        //Bundle bundle = new Bundle();
        //bundle.putString(Constants.TOAST, "Device connection was lost");
        //msg.setData(bundle);
        //btHandler.sendMessage(msg);
    }

    /**
     * This thread runs while attempting to make an outgoing connection
     * with a device. It runs straight through; the connection either
     * succeeds or fails.
     */
    private class ConnectThread extends Thread {
        private final BluetoothSocket btSocket;
        private final BluetoothDevice btDevice;

        public ConnectThread(BluetoothDevice device) {
            btDevice = device;
            ParcelUuid[] uuids = device.getUuids();
            BluetoothSocket tmp = null;

            // Get a BluetoothSocket for a connection with the
            // given BluetoothDevice
            try {
                btDevice.fetchUuidsWithSdp();
                tmp = btDevice.createRfcommSocketToServiceRecord(uuids[0].getUuid());
            } catch (IOException e) {
                Log.e(TAG, "Socket for UUID: " + uuids[0].toString() + "create() failed", e);
            }
            btSocket = tmp;
        }

        public void run() {
            Log.i(TAG, "BEGIN ConnectThread SocketType");

            // Always cancel discovery because it will slow down a connection
            btAdapter.cancelDiscovery();

            // Make a connection to the BluetoothSocket
            try {
                // This is a blocking call and will only return on a
                // successful connection or an exception
                btSocket.connect();
            } catch (IOException e) {
                // Close the socket
                try {
                    btSocket.close();
                } catch (IOException e2) {
                    Log.e(TAG, "unable to close() socket during connection failure", e2);
                }
                connectionFailed();
                return;
            }

            // Reset the ConnectThread because we're done
            synchronized (BluetoothService.this) {
                btConnectThread = null;
            }

            // Start the connected thread
            connected(btSocket, btDevice);
        }

        public void cancel() {
            try {
                btSocket.close();
            } catch (IOException e) {
                Log.e(TAG, "close() of connect socket failed", e);
            }
        }
    }

    /**
     * This thread runs during a connection with a remote device.
     * It handles all incoming and outgoing transmissions.
     */
    private class ConnectedThread extends Thread {
        private final BluetoothSocket btSocket;
        private final InputStream btInStream;
        private final OutputStream btOutStream;

        public ConnectedThread(BluetoothSocket socket) {
            Log.d(TAG, "create ConnectedThread" );
            btSocket = socket;
            InputStream tmpIn = null;
            OutputStream tmpOut = null;

            // Get the BluetoothSocket input and output streams
            try {
                tmpIn = socket.getInputStream();
                tmpOut = socket.getOutputStream();
            } catch (IOException e) {
                Log.e(TAG, "temp sockets not created", e);
            }

            btInStream = tmpIn;
            btOutStream = tmpOut;
        }

        public void run() {
            Log.i(TAG, "BEGIN ConnectedThread: state = " + btState);
            byte[] buffer = new byte[1024];
            List<Byte> toSend = new ArrayList<>();
            int bytes;

            // Keep listening to the InputStream while connected
            while (btState == STATE_CONNECTED) {
                try {
                    // Read from the InputStream
                    bytes = btInStream.read(buffer);
                    for (int i = 0; i < bytes; i++) {
                        if (buffer[i] == '\n') {
                            // Send the obtained bytes to the UI Activity
                            byte[] array = new byte[toSend.size()];
                            int j = 0;
                            for (Byte current : toSend) {
                                array[j] = current;
                                j++;
                            }
                            btHandler.obtainMessage(Constants.MESSAGE_READ, bytes, -1, new String(array))
                                    .sendToTarget();
                            toSend.clear();
                        } else {
                            toSend.add(buffer[i]);
                        }
                    }
                } catch (IOException e) {
                    Log.e(TAG, "disconnected", e);
                    //This check will help differentiate between a disconnection that was initiated
                    //and one that happened when we did not want to.
                    if (btState == STATE_CONNECTED ) {
                        connectionLost();
                    }
                    // Start the service over to restart listening mode
                    break;
                }
            }

        }
        /**
         * Write to the connected OutStream.
         *
         * @param buffer The bytes to write
         */
        public void write(byte[] buffer) {
            Log.e(TAG, "Write to bt adapter.");
            try {
                btOutStream.write(buffer);
                // Share the sent message back to the UI Activity
                btHandler.obtainMessage(Constants.MESSAGE_WRITE, -1, -1, buffer)
                        .sendToTarget();
            } catch (IOException e) {
                Log.e(TAG, "Exception during write", e);
            }
        }

        public void cancel() {
            Log.d(TAG, "Connected thread cancel.");
            try {
                btSocket.close();
            } catch (IOException e) {
                Log.e(TAG, "close() of connect socket failed", e);
            }
        }
    }
}


